{---------------------------------------------------------------------

  TntRZNEdit, a component with edit controls for
    real numbers      -> TTntEditReal
    integers          -> TTntEditInteger
    positive integers -> TTntEditCardinal
    
  This is my version of a numeric input component. It comes with three
  types, each handling its special input range ( real, integer and
  cardinal ).
  All input is watched, a TTntEditInteger accepts only characters in
  the set [-,0,1,..,9] for instance.
  The clipboard is supported.
  An empty input is interpreted as 0.
  For TTntEditReal and TTntEditInteger an input containing only a plus-sign
  or a minus-sign is also regarded as 0.
  Ancestor class is the TTntCustomEdit, new properties are:
    FormatString: is used in the FormatFloat-routine from Delphi
    MaxValue \
             : determine the input boundaries, if Validate is True  
    MinValue /
    Validate : determines if input is checked against boundaries
    Value    : the result
  This is my adaption of several numerical input controls, which I
  found on the Delphi Super Page. I give credits to DaveH, Arnt Kern,
  Richard Vowles and Robert Vivrette.
  rk 08.02.96
  Robert Kratz                                  kratz@labs.polycnrs-gre.fr
 
 (c) Copyleft or copyright under terms of GNU General Public License
 2003-2006  Jordi March i Nogu   jmarch@comg.es
 03/01/2006 version

----------------------------------------------------------------------}

unit TntRZNEdit;
{$D+,L+}

interface

uses
  WinTypes, Classes, StdCtrls, ExtCtrls, Controls, SysUtils,
  GrTypFun, TntRZNE0;

type
  TTntRZNSNumber = class(TTntRZNNumber)
  public
    function  Valid: Boolean; override;
  published
    property Color;
  end;

  TTntEditReal = class(TTntRZNSNumber)
  private
    FDecimalPlaces: Byte;
    FMinValue, FMaxValue: TVReal;
    function  CheckValue (NewValue: Real): Real;
    function  EmptyOrSign: Boolean;
    procedure SetDecimalPlaces (const Value: Byte);
    function  GetLimValue (Lim: TVReal): string;
    procedure SetLimValue (const V: string;  var Lim: TVReal);
    function  GetNMax: Real;
    procedure SetNMax (R: Real);
    function  GetNMin: Real;
    procedure SetNMin (R: Real);
    function  GetValue: Real;
    procedure SetValue (NewValue: Real);
  protected {for no compiler Hint message}
    procedure FormatValue; override;
    function  FFormatString: string; override;
    function  GetCMaxValue: string; override;
    function  GetCMinValue: string; override;
    procedure SetFormatString; override;
    procedure SetMaxValue (const V: string); override;
    procedure SetMinValue (const V: string); override;
    function  MaxValueAssig: Boolean; override;
    function  MinValueAssig: Boolean; override;
    procedure SetValidate (B:  Boolean); override;
    procedure SetValueValid; override;
    function  SyntaxOk: Boolean; override;
    function  ValidChars: WideString; override;
  public
    constructor Create (AOwner: TComponent); override;
    function  ValidNoMess: Boolean; override;
    property  NMax: Real read GetNMax write SetNMax;
    property  NMin: Real read GetNMin write SetNMin;
    property  NVMax: TVReal read FMaxValue write FMaxValue;
    property  NVMin: TVReal read FMinValue write FMinValue;
  published
    property DecimalPlaces: Byte read FDecimalPlaces write SetDecimalPlaces;
    property Value: Real read GetValue write SetValue;
  end;

  TTntEditInteger = class(TTntRZNSNumber)
  private
    FDecimalsAccept: Boolean;
    FMinValue, FMaxValue: TVLongint;
    function  CheckValue (NewValue: Longint): Longint;
    function  EmptyOrSign: Boolean;
    function  GetNMax: Longint;
    procedure SetNMax (R: Longint);
    function  GetNMin: Longint;
    procedure SetNMin (R: Longint);
    function  GetLimValue (Lim: TVLongint): string;
    procedure SetLimValue (const V: string;  var Lim: TVLongint);
    function  GetValue: Longint;
    procedure SetValue (NewValue: Longint);
  protected {for no compiler Hint message}
    procedure FormatValue; override;
    function  FFormatString: string; override;
    function  GetCMaxValue: string; override;
    function  GetCMinValue: string; override;
    procedure SetFormatString; override;
    procedure SetMaxValue (const V: string); override;
    procedure SetMinValue (const V: string); override;
    function  MaxValueAssig: Boolean; override;
    function  MinValueAssig: Boolean; override;
    function  SyntaxOk: Boolean; override;
    procedure SetValidate (B:  Boolean); override;
    procedure SetValueValid; override;
    function  ValidChars: WideString; override;
  public
    constructor Create (AOwner: TComponent); override;
    function  ValidNoMess: Boolean; override;
    property  NMax: Longint read GetNMax write SetNMax;
    property  NMin: Longint read GetNMin write SetNMin;
    property  NVMax: TVLongint read FMaxValue write FMaxValue;
    property  NVMin: TVLongint read FMinValue write FMinValue;
  published
    property DecimalsAccept: Boolean read FDecimalsAccept write FDecimalsAccept default False;
    property Value: Longint read GetValue write SetValue;
  end;

  TTntEditCardinal = class(TTntEditInteger)
  private
    procedure SetLimValue (const V: string;  var Lim: TVLongint);
  protected {for no compiler Hint message}
    procedure SetMaxValue (const V: string); override;
    procedure SetMinValue (const V: string); override;
    function  ValidChars: WideString; override;
  end;

implementation

uses
  WinProcs, Math,
  StrNumU, TntMessUsual;

{--------------------------------------}
{-PUBLIC-----------}

function TTntRZNSNumber.Valid: Boolean;
begin
  Result := ValidNoMess;
  if  not Result  then  begin
    WideMessageBoxWarning (WhenRangeErrorStr);
    SetFocus;
  end;
end {Valid};

{--------------------------------------}

function TTntEditReal.CheckValue (NewValue: Real): Real;
begin
  if FValidate
  then  CheckValueReal (NewValue, FMinValue, FMaxValue);
  Result := NewValue;
end {CheckValue};

function TTntEditReal.EmptyOrSign: Boolean;
begin
  Result := (VText='') or (VText='-') or (VText='+');
end {EmptyOrSign};

procedure TTntEditReal.SetDecimalPlaces (const Value: Byte);
begin
  if  Value <> FDecimalPlaces  then  begin
    FDecimalPlaces := Value;
    FormatValue;
  end;
end {SetDecimalPlaces};

function TTntEditReal.GetLimValue (Lim: TVReal): string;
begin
  if  Lim.Valid
  then  Result := FloatToStr (Lim.Num)
  else  Result := '';
end {GetLimValue};

procedure TTntEditReal.SetLimValue (const V: string;  var Lim: TVReal);
begin
  Lim.Valid := StrToFloatLim (V, Lim.Num);
  ValidNoMess;
end {SetLimValue};

function TTntEditReal.GetNMax: Real;
begin
  Result := GetFromVReal (FMaxValue);
end {GetNMax};

procedure TTntEditReal.SetNMax (R: Real);
begin
  SetToVReal (FMaxValue, R);
  ValidNoMess;
end {SetNMax};

function TTntEditReal.GetNMin: Real;
begin
  Result := GetFromVReal (FMinValue);
end {GetNMin};

procedure TTntEditReal.SetNMin (R: Real);
begin
  SetToVReal (FMinValue, R);
  ValidNoMess;
end {SetNMin};

function TTntEditReal.GetValue: Real;
begin
  if  not StrToFloat0 (Text, Result)  then  begin
    if  FMinValue.Valid
    then  Result := FMinValue.Num
    else  Result := 0;
  end;
end {GetValue};

procedure TTntEditReal.SetValue (NewValue: Real);
var
  R: Real;
begin
  try
    R := CheckValue (NewValue);
    if  (R <> NewValue) or (not NotFormat)
    then  Text := FormatFloat (FFormatString, R);
  except
  end;
  ValidNoMess;
end {SetValue};

{-PROTECTED--------}

procedure TTntEditReal.FormatValue;
begin
  Text := FormatFloat (FFormatString, Value);
end {FormatValue};

function TTntEditReal.FFormatString: string;
begin
  Result := FloatFormat (FWithThousandSeparator, FDecimalPlaces);
end {FFormatString};

function TTntEditReal.GetCMaxValue: string;
begin
  Result := GetLimValue (FMaxValue);
end {GetCMaxValue};

function TTntEditReal.GetCMinValue: string;
begin
  Result := GetLimValue (FMinValue);
end {GetCMinValue};

procedure TTntEditReal.SetFormatString;
begin
  SetValue (Value);
end {SetFormatString};

procedure TTntEditReal.SetMaxValue (const V: string);
begin
  SetLimValue (V, FMaxValue);
end {SetMaxValue};

procedure TTntEditReal.SetMinValue (const V: string);
begin
  SetLimValue (V, FMinValue);
end {SetMinValue};

function TTntEditReal.MaxValueAssig: Boolean;
begin
  MaxValueAssig := FMaxValue.Valid;
end {MaxValueAssig};

function TTntEditReal.MinValueAssig: Boolean;
begin
  MinValueAssig := FMinValue.Valid;
end {MinValueAssig};

procedure TTntEditReal.SetValidate (B:  Boolean);
begin
  FValidate := B;
  SetValue (Value);
end {SetValidate};

procedure TTntEditReal.SetValueValid;
begin
  SetValue (Value);
end {SetValueValid};

function TTntEditReal.SyntaxOk: Boolean;
begin
  try
    {-----------------------------------------------------------
      Syntax correct if transformation causes no exception.
      if the input is empty or a sign ( VText = '', '+', '-' ),
      it is interpreted as 0
     -----------------------------------------------------------}
    if not EmptyOrSign then StrToFloat (VText);
    Result := True;
  except
    on EConvertError do Result := False;
  end;
end {SyntaxOk};

function TTntEditReal.ValidChars: WideString;
begin
  ValidChars := SetOfRealChars;
end {ValidChars};

{-PUBLIC-----------}

constructor TTntEditReal.Create (AOwner: TComponent);
begin
  inherited Create (AOwner);
  CharCase := ecUpperCase;
  FMinValue := VRealNul;
  FMaxValue := VRealNul;
  SetValue (0);
end {Create};

function TTntEditReal.ValidNoMess: Boolean;
var
  IsValid: Boolean;
begin
  IsValid := RealValueValid (Value, FMinValue, FMaxValue);
  SetFontColorValid (IsValid);
  Result := IsValid;
end {ValidNoMess};

{--------------------------------------}
{-PRIVATE----------}

function TTntEditInteger.CheckValue (NewValue: Longint): Longint;
begin
  if FValidate
  then  CheckValueLongint (NewValue, FMinValue, FMaxValue);
  Result := NewValue;
end {CheckValue};

function TTntEditInteger.EmptyOrSign: Boolean;
begin
  Result := (VText='') or (VText='-') or (VText='+');
end {EmptyOrSign};

function TTntEditInteger.GetNMax: Longint;
begin
  Result := GetFromVLongint (FMaxValue);
end {GetNMax};

procedure TTntEditInteger.SetNMax (R: Longint);
begin
  SetToVLongint (FMaxValue, R);
  ValidNoMess;
end {SetNMax};

function TTntEditInteger.GetNMin: Longint;
begin
  Result := GetFromVLongint (FMinValue);
end {GetNMin};

procedure TTntEditInteger.SetNMin (R: Longint);
begin
  SetToVLongint (FMinValue, R);
  ValidNoMess;
end {SetNMin};

function TTntEditInteger.GetLimValue (Lim: TVLongInt): string;
begin
  if  Lim.Valid
  then  Result := IntToStr (Lim.Num)
  else  Result := '';
end {GetLimValue};

procedure TTntEditInteger.SetLimValue (const V: string;  var Lim: TVLongInt);
begin
  Lim.Valid := StrToInt0 (V, Lim.Num);
  ValidNoMess;
end {SetLimValue};

function TTntEditInteger.GetValue: Longint;
var
  VValue: TVLongint;
begin
  VValue := StrToVValue (Text, FDecimalsAccept);
  if VValue.Valid
  then
    Result := GetFromVLongint (VValue)
  else  begin
    if  FMinValue.Valid
    then  Result := FMinValue.Num
    else  Result := 0;
  end;
end {GetValue};

procedure TTntEditInteger.SetValue (NewValue: Longint);
var
  LI: Longint;
begin
  try
    LI := CheckValue (NewValue);
    if  (LI <> NewValue) or (not NotFormat)
    then  Text := FormatFloat (FFormatString, LI);
  except
  end;
  ValidNoMess;
end {SetValue};

{-PROTECTED--------}

procedure TTntEditInteger.FormatValue;
begin
  Text := FormatFloat (FFormatString, Value);
end {FormatValue};

function TTntEditInteger.FFormatString: string;
begin
  Result := IntegerFormat (FWithThousandSeparator);
end {FFormatString};

function TTntEditInteger.GetCMaxValue: string;
begin
  Result := GetLimValue (FMaxValue);
end {GetCMaxValue};

function TTntEditInteger.GetCMinValue: string;
begin
  Result := GetLimValue (FMinValue);
end {GetCMinValue};

procedure TTntEditInteger.SetFormatString;
begin
  SetValue (Value);
end {SetFormatString};

procedure TTntEditInteger.SetMaxValue (const V: string);
begin
  SetLimValue (V, FMaxValue);
end {SetMaxValue};

procedure TTntEditInteger.SetMinValue (const V: string);
begin
  SetLimValue (V, FMinValue);
end {SetMinValue};

function TTntEditInteger.MaxValueAssig: Boolean;
begin
  MaxValueAssig := FMaxValue.Valid;
end {MaxValueAssig};

function TTntEditInteger.MinValueAssig: Boolean;
begin
  MinValueAssig := FMinValue.Valid;
end {MinValueAssig};

function TTntEditInteger.SyntaxOk: Boolean;
begin
  try
    {-----------------------------------------------------------
      Syntax correct if transformation causes no exception.
      if the input is empty or a sign  (VText = '', '+', '-'),
      it is interpreted as 0
     -----------------------------------------------------------}
    if not EmptyOrSign then begin
      if  FDecimalsAccept and (Pos(DecimalSeparator,VText)<>0)
      then  StrToFloat(VText)
      else  StrToInt(VText);
    end;
    Result := True;
  except
    on EConvertError do Result := False;
  end;
end {SyntaxOk};

procedure TTntEditInteger.SetValidate (B:  Boolean);
begin
  FValidate := B;
  SetValue (Value);
end {SetValidate};

procedure TTntEditInteger.SetValueValid;
begin
  SetValue (Value);
end {SetValueValid};

function TTntEditInteger.ValidChars: WideString;
begin
  ValidChars := SetOfIntegerChars (FDecimalsAccept);
end {ValidChars};

{-PUBLIC-----------}

constructor TTntEditInteger.Create (AOwner: TComponent);
begin
  inherited Create (AOwner);
  FMinValue := VLongintNul;
  FMaxValue := VLongintNul;
  SetValue (0);
end {Create};

function TTntEditInteger.ValidNoMess: Boolean;
var
  IsValid: Boolean;
begin
  IsValid := LongintValueValid (Value, FMinValue, FMaxValue);
  SetFontColorValid (IsValid);
  Result := IsValid;
end {ValidNoMess};

{--------------------------------------------------------------------------}
{-PRIVATE----------}

procedure TTntEditCardinal.SetLimValue (const V: string;  var Lim: TVLongint);
begin
  Lim.Valid := StrToInt0 (V, Lim.Num);
  if  Lim.Valid
  then  Lim.Num := Max (Lim.Num, 0);
  ValidNoMess;
end {SetLimValue};

{-PROTECTED--------}

procedure TTntEditCardinal.SetMaxValue (const V: string);
begin
  SetLimValue (V, FMaxValue);
end {SetMaxValue};

procedure TTntEditCardinal.SetMinValue (const V: string);
begin
  SetLimValue (V, FMinValue);
end {SetMinValue};

function TTntEditCardinal.ValidChars: WideString;
begin
  ValidChars := SetOfCardinalChars (FDecimalsAccept);
end {ValidChars};

{--------------------------------------}

end.
